@@ -20,18 +20,19 @@
 #include "Compiler.hh"
 #include "Exception.hh"
 
+#include <random>
 #include <sstream>
 
-#include <boost/crc.hpp> // for boost::crc_32_type
-#include <boost/iostreams/device/file.hpp>
-#include <boost/iostreams/filter/gzip.hpp>
-#include <boost/iostreams/filter/zlib.hpp>
-#include <boost/random/mersenne_twister.hpp>
-
 #ifdef SNAPPY_CODEC_AVAILABLE
 #include <snappy.h>
 #endif
 
+#ifdef ZSTD_CODEC_AVAILABLE
+#include <zstd.h>
+#endif
+
+#include <zlib.h>
+
 namespace avro {
 using std::copy;
 using std::istringstream;
@@ -52,50 +53,54 @@ const string AVRO_DEFLATE_CODEC("deflate");
 const string AVRO_SNAPPY_CODEC = "snappy";
 #endif
 
+#ifdef ZSTD_CODEC_AVAILABLE
+const string AVRO_ZSTD_CODEC = "zstandard";
+#endif
+
 const size_t minSyncInterval = 32;
 const size_t maxSyncInterval = 1u << 30;
 
-boost::iostreams::zlib_params get_zlib_params() {
-    boost::iostreams::zlib_params ret;
-    ret.method = boost::iostreams::zlib::deflated;
-    ret.noheader = true;
-    return ret;
-}
+// Recommended by https://www.zlib.net/zlib_how.html
+const size_t zlibBufGrowSize = 128 * 1024;
+
 } // namespace
 
 DataFileWriterBase::DataFileWriterBase(const char *filename, const ValidSchema &schema, size_t syncInterval,
-                                       Codec codec) : filename_(filename),
-                                                      schema_(schema),
-                                                      encoderPtr_(binaryEncoder()),
-                                                      syncInterval_(syncInterval),
-                                                      codec_(codec),
-                                                      stream_(fileOutputStream(filename)),
-                                                      buffer_(memoryOutputStream()),
-                                                      sync_(makeSync()),
-                                                      objectCount_(0),
-                                                      lastSync_(0) {
+                                       Codec codec, const Metadata &metadata) : filename_(filename),
+                                                                                schema_(schema),
+                                                                                encoderPtr_(binaryEncoder()),
+                                                                                syncInterval_(syncInterval),
+                                                                                codec_(codec),
+                                                                                stream_(fileOutputStream(filename)),
+                                                                                buffer_(memoryOutputStream()),
+                                                                                sync_(makeSync()),
+                                                                                objectCount_(0),
+                                                                                metadata_(metadata),
+                                                                                lastSync_(0) {
     init(schema, syncInterval, codec);
 }
 
 DataFileWriterBase::DataFileWriterBase(std::unique_ptr<OutputStream> outputStream,
-                                       const ValidSchema &schema, size_t syncInterval, Codec codec) : filename_(),
-                                                                                                      schema_(schema),
-                                                                                                      encoderPtr_(binaryEncoder()),
-                                                                                                      syncInterval_(syncInterval),
-                                                                                                      codec_(codec),
-                                                                                                      stream_(std::move(outputStream)),
-                                                                                                      buffer_(memoryOutputStream()),
-                                                                                                      sync_(makeSync()),
-                                                                                                      objectCount_(0),
-                                                                                                      lastSync_(0) {
+                                       const ValidSchema &schema, size_t syncInterval,
+                                       Codec codec, const Metadata &metadata) : filename_(),
+                                                                                schema_(schema),
+                                                                                encoderPtr_(binaryEncoder()),
+                                                                                syncInterval_(syncInterval),
+                                                                                codec_(codec),
+                                                                                stream_(std::move(outputStream)),
+                                                                                buffer_(memoryOutputStream()),
+                                                                                sync_(makeSync()),
+                                                                                objectCount_(0),
+                                                                                metadata_(metadata),
+                                                                                lastSync_(0) {
     init(schema, syncInterval, codec);
 }
 
 void DataFileWriterBase::init(const ValidSchema &schema, size_t syncInterval, const Codec &codec) {
     if (syncInterval < minSyncInterval || syncInterval > maxSyncInterval) {
-        throw Exception(boost::format("Invalid sync interval: %1%. "
-                                      "Should be between %2% and %3%")
-                        % syncInterval % minSyncInterval % maxSyncInterval);
+        throw Exception(
+            "Invalid sync interval: {}. Should be between {} and {}",
+            syncInterval, minSyncInterval, maxSyncInterval);
     }
     setMetadata(AVRO_CODEC_KEY, AVRO_NULL_CODEC);
 
@@ -106,9 +111,13 @@ void DataFileWriterBase::init(const ValidSchema &schema, size_t syncInterval, co
 #ifdef SNAPPY_CODEC_AVAILABLE
     } else if (codec_ == SNAPPY_CODEC) {
         setMetadata(AVRO_CODEC_KEY, AVRO_SNAPPY_CODEC);
+#endif
+#ifdef ZSTD_CODEC_AVAILABLE
+    } else if (codec_ == ZSTD_CODEC) {
+        setMetadata(AVRO_CODEC_KEY, AVRO_ZSTD_CODEC);
 #endif
     } else {
-        throw Exception(boost::format("Unknown codec: %1%") % codec);
+        throw Exception("Unknown codec: {}", int(codec));
     }
     setMetadata(AVRO_SCHEMA_KEY, schema.toJson(false));
 
@@ -120,7 +129,9 @@ void DataFileWriterBase::init(const ValidSchema &schema, size_t syncInterval, co
 
 DataFileWriterBase::~DataFileWriterBase() {
     if (stream_) {
-        close();
+        try {
+            close();
+        } catch (...) {}
     }
 }
 
@@ -141,21 +152,45 @@ void DataFileWriterBase::sync() {
         std::unique_ptr<InputStream> in = memoryInputStream(*buffer_);
         copy(*in, *stream_);
     } else if (codec_ == DEFLATE_CODEC) {
-        std::vector<char> buf;
+        std::vector<uint8_t> buf;
         {
-            boost::iostreams::filtering_ostream os;
-            os.push(boost::iostreams::zlib_compressor(get_zlib_params()));
-            os.push(boost::iostreams::back_inserter(buf));
-            const uint8_t *data;
-            size_t len;
+            z_stream zs;
+            zs.zalloc = Z_NULL;
+            zs.zfree = Z_NULL;
+            zs.opaque = Z_NULL;
+
+            int ret = deflateInit2(&zs, Z_DEFAULT_COMPRESSION, Z_DEFLATED, -15, 8, Z_DEFAULT_STRATEGY);
+            if (ret != Z_OK) {
+                throw Exception("Failed to initialize deflate, error: {}", ret);
+            }
 
             std::unique_ptr<InputStream> input = memoryInputStream(*buffer_);
-            while (input->next(&data, &len)) {
-                boost::iostreams::write(os, reinterpret_cast<const char *>(data), len);
+            const uint8_t *data;
+            size_t len;
+            while (ret != Z_STREAM_END && input->next(&data, &len)) {
+                zs.avail_in = static_cast<uInt>(len);
+                zs.next_in = const_cast<Bytef *>(data);
+                bool flush = (zs.total_in + len) >= buffer_->byteCount();
+                do {
+                    if (zs.total_out == buf.size()) {
+                        buf.resize(buf.size() + zlibBufGrowSize);
+                    }
+                    zs.avail_out = static_cast<uInt>(buf.size() - zs.total_out);
+                    zs.next_out = buf.data() + zs.total_out;
+                    ret = deflate(&zs, flush ? Z_FINISH : Z_NO_FLUSH);
+                    if (ret == Z_STREAM_END) {
+                        break;
+                    }
+                    if (ret != Z_OK) {
+                        throw Exception("Failed to deflate, error: {}", ret);
+                    }
+                } while (zs.avail_out == 0);
             }
+
+            buf.resize(zs.total_out);
+            (void) deflateEnd(&zs);
         } // make sure all is flushed
-        std::unique_ptr<InputStream> in = memoryInputStream(
-            reinterpret_cast<const uint8_t *>(buf.data()), buf.size());
+        std::unique_ptr<InputStream> in = memoryInputStream(buf.data(), buf.size());
         int64_t byteCount = buf.size();
         avro::encode(*encoderPtr_, byteCount);
         encoderPtr_->flush();
@@ -164,45 +199,71 @@ void DataFileWriterBase::sync() {
     } else if (codec_ == SNAPPY_CODEC) {
         std::vector<char> temp;
         std::string compressed;
-        boost::crc_32_type crc;
-        {
-            boost::iostreams::filtering_ostream os;
-            os.push(boost::iostreams::back_inserter(temp));
-            const uint8_t *data;
-            size_t len;
 
-            std::unique_ptr<InputStream> input = memoryInputStream(*buffer_);
-            while (input->next(&data, &len)) {
-                boost::iostreams::write(os, reinterpret_cast<const char *>(data),
-                                        len);
-            }
-        } // make sure all is flushed
+        const uint8_t *data;
+        size_t len;
+        std::unique_ptr<InputStream> input = memoryInputStream(*buffer_);
+        while (input->next(&data, &len)) {
+            temp.insert(temp.end(), reinterpret_cast<const char *>(data),
+                        reinterpret_cast<const char *>(data) + len);
+        }
 
-        crc.process_bytes(reinterpret_cast<const char *>(temp.data()),
-                          temp.size());
         // For Snappy, add the CRC32 checksum
-        int32_t checksum = crc();
+        auto checksum = crc32(0, reinterpret_cast<const Bytef *>(temp.data()),
+                              static_cast<uInt>(temp.size()));
 
         // Now compress
         size_t compressed_size = snappy::Compress(
             reinterpret_cast<const char *>(temp.data()), temp.size(),
             &compressed);
+
         temp.clear();
-        {
-            boost::iostreams::filtering_ostream os;
-            os.push(boost::iostreams::back_inserter(temp));
-            boost::iostreams::write(os, compressed.c_str(), compressed_size);
-        }
-        temp.push_back((checksum >> 24) & 0xFF);
-        temp.push_back((checksum >> 16) & 0xFF);
-        temp.push_back((checksum >> 8) & 0xFF);
-        temp.push_back(checksum & 0xFF);
+        temp.insert(temp.end(), compressed.c_str(),
+                    compressed.c_str() + compressed_size);
+
+        temp.push_back(static_cast<char>((checksum >> 24) & 0xFF));
+        temp.push_back(static_cast<char>((checksum >> 16) & 0xFF));
+        temp.push_back(static_cast<char>((checksum >> 8) & 0xFF));
+        temp.push_back(static_cast<char>(checksum & 0xFF));
         std::unique_ptr<InputStream> in = memoryInputStream(
             reinterpret_cast<const uint8_t *>(temp.data()), temp.size());
         int64_t byteCount = temp.size();
         avro::encode(*encoderPtr_, byteCount);
         encoderPtr_->flush();
         copy(*in, *stream_);
+#endif
+#ifdef ZSTD_CODEC_AVAILABLE
+    } else if (codec_ == ZSTD_CODEC) {
+        // Read all uncompressed data into a single buffer
+        std::vector<char> uncompressed;
+        const uint8_t *data;
+        size_t len;
+        std::unique_ptr<InputStream> input = memoryInputStream(*buffer_);
+        while (input->next(&data, &len)) {
+            uncompressed.insert(uncompressed.end(), reinterpret_cast<const char *>(data),
+                                reinterpret_cast<const char *>(data) + len);
+        }
+
+        // Pre-allocate buffer for compressed data
+        size_t max_compressed_size = ZSTD_compressBound(uncompressed.size());
+        std::vector<char> compressed(max_compressed_size);
+
+        // Compress the data using ZSTD block API
+        size_t compressed_size = ZSTD_compress(
+            compressed.data(), max_compressed_size,
+            uncompressed.data(), uncompressed.size(),
+            ZSTD_CLEVEL_DEFAULT);
+
+        if (ZSTD_isError(compressed_size)) {
+            throw Exception("ZSTD compression error: {}", ZSTD_getErrorName(compressed_size));
+        }
+
+        compressed.resize(compressed_size);
+        std::unique_ptr<InputStream> in = memoryInputStream(
+            reinterpret_cast<const uint8_t *>(compressed.data()), compressed.size());
+        avro::encode(*encoderPtr_, static_cast<int64_t>(compressed_size));
+        encoderPtr_->flush();
+        copy(*in, *stream_);
 #endif
     }
 
@@ -233,7 +294,7 @@ void DataFileWriterBase::flush() {
 }
 
 DataFileSync DataFileWriterBase::makeSync() {
-    boost::mt19937 random(static_cast<uint32_t>(time(nullptr)));
+    std::mt19937 random(static_cast<uint32_t>(time(nullptr)));
     DataFileSync sync;
     std::generate(sync.begin(), sync.end(), random);
     return sync;
@@ -256,14 +317,14 @@ void DataFileWriterBase::setMetadata(const string &key, const string &value) {
     metadata_[key] = v;
 }
 
-DataFileReaderBase::DataFileReaderBase(const char *filename) : filename_(filename), codec_(NULL_CODEC), stream_(fileSeekableInputStream(filename)),
-                                                               decoder_(binaryDecoder()), objectCount_(0), eof_(false), blockStart_(-1),
-                                                               blockEnd_(-1) {
+DataFileReaderBase::DataFileReaderBase(const char *filename) : filename_(filename), stream_(fileSeekableInputStream(filename)),
+                                                               decoder_(binaryDecoder()), objectCount_(0), eof_(false),
+                                                               codec_(NULL_CODEC), blockStart_(-1), blockEnd_(-1) {
     readHeader();
 }
 
-DataFileReaderBase::DataFileReaderBase(std::unique_ptr<InputStream> inputStream) : codec_(NULL_CODEC), stream_(std::move(inputStream)),
-                                                                                   decoder_(binaryDecoder()), objectCount_(0), eof_(false) {
+DataFileReaderBase::DataFileReaderBase(std::unique_ptr<InputStream> inputStream) : stream_(std::move(inputStream)),
+                                                                                   decoder_(binaryDecoder()), objectCount_(0), eof_(false), codec_(NULL_CODEC) {
     readHeader();
 }
 
@@ -282,8 +343,7 @@ void DataFileReaderBase::init(const ValidSchema &readerSchema) {
 static void drain(InputStream &in) {
     const uint8_t *p = nullptr;
     size_t n = 0;
-    while (in.next(&p, &n))
-        ;
+    while (in.next(&p, &n));
 }
 
 char hex(unsigned int x) {
@@ -381,7 +441,6 @@ void DataFileReaderBase::readDataBlock() {
         dataStream_ = std::move(st);
 #ifdef SNAPPY_CODEC_AVAILABLE
     } else if (codec_ == SNAPPY_CODEC) {
-        boost::crc_32_type crc;
         uint32_t checksum = 0;
         compressed_.clear();
         uncompressed.clear();
@@ -405,41 +464,118 @@ void DataFileReaderBase::readDataBlock() {
             throw Exception(
                 "Snappy Compression reported an error when decompressing");
         }
-        crc.process_bytes(uncompressed.c_str(), uncompressed.size());
-        uint32_t c = crc();
+        auto c = crc32(0, reinterpret_cast<const Bytef *>(uncompressed.c_str()),
+                       static_cast<uInt>(uncompressed.size()));
         if (checksum != c) {
             throw Exception(
-                boost::format("Checksum did not match for Snappy compression: Expected: %1%, computed: %2%") % checksum
-                % c);
+                "Checksum did not match for Snappy compression: Expected: {}, computed: {}",
+                checksum, c);
         }
-        os_.reset(new boost::iostreams::filtering_istream());
-        os_->push(
-            boost::iostreams::basic_array_source<char>(uncompressed.c_str(),
-                                                       uncompressed.size()));
-        std::unique_ptr<InputStream> in = istreamInputStream(*os_);
+
+        std::unique_ptr<InputStream> in = memoryInputStream(
+            reinterpret_cast<const uint8_t *>(uncompressed.c_str()),
+            uncompressed.size());
 
         dataDecoder_->init(*in);
         dataStream_ = std::move(in);
 #endif
-    } else {
+#ifdef ZSTD_CODEC_AVAILABLE
+    } else if (codec_ == ZSTD_CODEC) {
         compressed_.clear();
         const uint8_t *data;
         size_t len;
         while (st->next(&data, &len)) {
             compressed_.insert(compressed_.end(), data, data + len);
         }
-        os_.reset(new boost::iostreams::filtering_istream());
-        os_->push(boost::iostreams::zlib_decompressor(get_zlib_params()));
-        os_->push(boost::iostreams::basic_array_source<char>(
-            compressed_.data(), compressed_.size()));
 
-        std::unique_ptr<InputStream> in = nonSeekableIstreamInputStream(*os_);
+        // Get the decompressed size
+        size_t decompressed_size = ZSTD_getFrameContentSize(
+            reinterpret_cast<const char *>(compressed_.data()), compressed_.size());
+        if (decompressed_size == ZSTD_CONTENTSIZE_ERROR) {
+            throw Exception("ZSTD: Not a valid compressed frame");
+        } else if (decompressed_size == ZSTD_CONTENTSIZE_UNKNOWN) {
+            throw Exception("ZSTD: Unable to determine decompressed size");
+        }
+
+        // Decompress the data
+        uncompressed.clear();
+        uncompressed.resize(decompressed_size);
+        size_t result = ZSTD_decompress(
+            uncompressed.data(), decompressed_size,
+            reinterpret_cast<const char *>(compressed_.data()), compressed_.size());
+
+        if (ZSTD_isError(result)) {
+            throw Exception("ZSTD decompression error: {}", ZSTD_getErrorName(result));
+        }
+        if (result != decompressed_size) {
+            throw Exception("ZSTD: Decompressed size mismatch: expected {}, got {}",
+                            decompressed_size, result);
+        }
+
+        std::unique_ptr<InputStream> in = memoryInputStream(
+            reinterpret_cast<const uint8_t *>(uncompressed.data()),
+            uncompressed.size());
+
+        dataDecoder_->init(*in);
+        dataStream_ = std::move(in);
+#endif
+    } else {
+        compressed_.clear();
+        uncompressed.clear();
+
+        {
+            z_stream zs;
+            zs.zalloc = Z_NULL;
+            zs.zfree = Z_NULL;
+            zs.opaque = Z_NULL;
+            zs.avail_in = 0;
+            zs.next_in = Z_NULL;
+
+            int ret = inflateInit2(&zs, /*windowBits=*/-15);
+            if (ret != Z_OK) {
+                throw Exception("Failed to initialize inflate, error: {}", ret);
+            }
+
+            const uint8_t *data;
+            size_t len;
+            while (ret != Z_STREAM_END && st->next(&data, &len)) {
+                zs.avail_in = static_cast<uInt>(len);
+                zs.next_in = const_cast<Bytef *>(data);
+                do {
+                    if (zs.total_out == uncompressed.size()) {
+                        uncompressed.resize(uncompressed.size() + zlibBufGrowSize);
+                    }
+                    zs.avail_out = static_cast<uInt>(uncompressed.size() - zs.total_out);
+                    zs.next_out = reinterpret_cast<Bytef *>(uncompressed.data() + zs.total_out);
+                    ret = inflate(&zs, Z_NO_FLUSH);
+                    if (ret == Z_STREAM_END) {
+                        break;
+                    }
+                    if (ret != Z_OK) {
+                        throw Exception("Failed to inflate, error: {}", ret);
+                    }
+                } while (zs.avail_out == 0);
+            }
+
+            uncompressed.resize(zs.total_out);
+            (void) inflateEnd(&zs);
+        }
+
+        std::unique_ptr<InputStream> in = memoryInputStream(
+            reinterpret_cast<const uint8_t *>(uncompressed.c_str()),
+            uncompressed.size());
+
         dataDecoder_->init(*in);
         dataStream_ = std::move(in);
     }
 }
 
 void DataFileReaderBase::close() {
+    stream_.reset();
+    eof_ = true;
+    objectCount_ = 0;
+    blockStart_ = 0;
+    blockEnd_ = 0;
 }
 
 static string toString(const vector<uint8_t> &v) {
@@ -453,7 +589,7 @@ static ValidSchema makeSchema(const vector<uint8_t> &v) {
     istringstream iss(toString(v));
     ValidSchema vs;
     compileJsonSchema(iss, vs);
-    return ValidSchema(vs);
+    return vs;
 }
 
 void DataFileReaderBase::readHeader() {
@@ -482,6 +618,11 @@ void DataFileReaderBase::readHeader() {
     } else if (it != metadata_.end()
                && toString(it->second) == AVRO_SNAPPY_CODEC) {
         codec_ = SNAPPY_CODEC;
+#endif
+#ifdef ZSTD_CODEC_AVAILABLE
+    } else if (it != metadata_.end()
+               && toString(it->second) == AVRO_ZSTD_CODEC) {
+        codec_ = ZSTD_CODEC;
 #endif
     } else {
         codec_ = NULL_CODEC;
@@ -525,8 +666,7 @@ void DataFileReaderBase::sync(int64_t position) {
             eof_ = true;
             return;
         }
-        int len =
-            std::min(static_cast<size_t>(SyncSize - i), n);
+        size_t len = std::min(SyncSize - i, n);
         memcpy(&sync_buffer[i], p, len);
         p += len;
         n -= len;
