@@ -16,24 +16,20 @@
  * limitations under the License.
  */
 
+#include <algorithm>
 #include <cctype>
 #ifndef _WIN32
 #include <ctime>
 #endif
 #include <fstream>
 #include <iostream>
+#include <locale>
 #include <map>
+#include <optional>
+#include <random>
 #include <set>
-
-#include <boost/algorithm/string.hpp>
-#include <boost/lexical_cast.hpp>
-#include <boost/program_options.hpp>
-
-#include <boost/random/mersenne_twister.hpp>
 #include <utility>
 
-#include <boost/algorithm/string_regex.hpp>
-
 #include "Compiler.hh"
 #include "NodeImpl.hh"
 #include "ValidSchema.hh"
@@ -48,17 +44,9 @@ using std::set;
 using std::string;
 using std::vector;
 
-using boost::lexical_cast;
-
 using avro::compileJsonSchema;
 using avro::ValidSchema;
 
-#if __cplusplus >= 201703L
-#define ANY_NS "std"
-#else
-#define ANY_NS "boost"
-#endif
-
 struct PendingSetterGetter {
     string structName;
     string type;
@@ -75,8 +63,22 @@ struct PendingConstructor {
     PendingConstructor(string sn, string n, bool im) : structName(std::move(sn)), memberName(std::move(n)), initMember(im) {}
 };
 
+class UnionCodeTracker {
+    std::string schemaFile_;
+    size_t unionNumber_ = 0;
+    std::map<std::vector<std::string>, std::string> unionBranchNameMapping_;
+    std::set<std::string> generatedUnionTraits_;
+
+public:
+    explicit UnionCodeTracker(const std::string &schemaFile);
+    std::optional<std::string> getExistingUnionName(const std::vector<std::string> &unionBranches) const;
+    std::string generateNewUnionName(const std::vector<std::string> &unionBranches);
+    bool unionTraitsAlreadyGenerated(const std::string &unionClassName) const;
+    void setTraitsGenerated(const std::string &unionClassName);
+};
+
 class CodeGen {
-    size_t unionNumber_;
+    UnionCodeTracker unionTracker_;
     std::ostream &os_;
     bool inNamespace_;
     const std::string ns_;
@@ -85,7 +87,7 @@ class CodeGen {
     const std::string includePrefix_;
     const bool noUnion_;
     const std::string guardString_;
-    boost::mt19937 random_;
+    std::mt19937 random_;
 
     vector<PendingSetterGetter> pendingGettersAndSetters;
     vector<PendingConstructor> pendingConstructors;
@@ -98,7 +100,6 @@ class CodeGen {
     std::string generateEnumType(const NodePtr &n);
     std::string cppTypeOf(const NodePtr &n);
     std::string generateRecordType(const NodePtr &n);
-    std::string unionName();
     std::string generateUnionType(const NodePtr &n);
     std::string generateType(const NodePtr &n);
     std::string generateDeclaration(const NodePtr &n);
@@ -107,17 +108,21 @@ class CodeGen {
     void generateTraits(const NodePtr &n);
     void generateRecordTraits(const NodePtr &n);
     void generateUnionTraits(const NodePtr &n);
+    void generateDocComment(const NodePtr &n, const char *indent = "");
     void emitCopyright();
+    void emitGeneratedWarning();
 
 public:
     CodeGen(std::ostream &os, std::string ns,
             std::string schemaFile, std::string headerFile,
             std::string guardString,
-            std::string includePrefix, bool noUnion) : unionNumber_(0), os_(os), inNamespace_(false), ns_(std::move(ns)),
+            std::string includePrefix, bool noUnion) : unionTracker_(schemaFile), os_(os), inNamespace_(false), ns_(std::move(ns)),
                                                        schemaFile_(std::move(schemaFile)), headerFile_(std::move(headerFile)),
                                                        includePrefix_(std::move(includePrefix)), noUnion_(noUnion),
                                                        guardString_(std::move(guardString)),
-                                                       random_(static_cast<uint32_t>(::time(nullptr))) {}
+                                                       random_(static_cast<uint32_t>(::time(nullptr))) {
+    }
+
     void generate(const ValidSchema &schema);
 };
 
@@ -125,7 +130,7 @@ static string decorate(const std::string &name) {
     static const char *cppReservedWords[] = {
         "alignas", "alignof", "and", "and_eq", "asm", "auto", "bitand", "bitor", "bool", "break",
         "case", "catch", "char", "char8_t", "char16_t", "char32_t", "class", "compl", "concept",
-        "const", "consteval", "constexpr", "const_cast", "continue", "co_await", "co_return",
+        "const", "consteval", "constexpr", "constinit", "const_cast", "continue", "co_await", "co_return",
         "co_yield", "decltype", "default", "delete", "do", "double", "dynamic_cast", "else",
         "enum", "explicit", "export", "extern", "false", "float", "for", "friend", "goto", "if",
         "import", "inline", "int", "long", "module", "mutable", "namespace", "new", "noexcept", "not",
@@ -186,7 +191,7 @@ string CodeGen::cppTypeOf(const NodePtr &n) {
         case avro::AVRO_MAP:
             return "std::map<std::string, " + cppTypeOf(n->leafAt(1)) + " >";
         case avro::AVRO_FIXED:
-            return "std::array<uint8_t, " + lexical_cast<string>(n->fixedSize()) + ">";
+            return "std::array<uint8_t, " + std::to_string(n->fixedSize()) + ">";
         case avro::AVRO_SYMBOLIC:
             return cppTypeOf(resolveSymbol(n));
         case avro::AVRO_UNION:
@@ -244,12 +249,18 @@ string CodeGen::generateRecordType(const NodePtr &n) {
         return it->second;
     }
 
+    generateDocComment(n);
     os_ << "struct " << decoratedName << " {\n";
     if (!noUnion_) {
         for (size_t i = 0; i < c; ++i) {
             if (n->leafAt(i)->type() == avro::AVRO_UNION) {
                 os_ << "    typedef " << types[i]
                     << ' ' << n->nameAt(i) << "_t;\n";
+                types[i] = n->nameAt(i) + "_t";
+            }
+            if (n->leafAt(i)->type() == avro::AVRO_ARRAY && n->leafAt(i)->leafAt(0)->type() == avro::AVRO_UNION) {
+                os_ << "    typedef " << types[i] << "::value_type"
+                    << ' ' << n->nameAt(i) << "_item_t;\n";
             }
         }
     }
@@ -257,11 +268,8 @@ string CodeGen::generateRecordType(const NodePtr &n) {
         // the nameAt(i) does not take c++ reserved words into account
         // so we need to call decorate on it
         std::string decoratedNameAt = decorate(n->nameAt(i));
-        if (!noUnion_ && n->leafAt(i)->type() == avro::AVRO_UNION) {
-            os_ << "    " << decoratedNameAt << "_t";
-        } else {
-            os_ << "    " << types[i];
-        }
+        generateDocComment(n->leafAt(i), "    ");
+        os_ << "    " << types[i];
         os_ << ' ' << decoratedNameAt << ";\n";
     }
 
@@ -275,13 +283,7 @@ string CodeGen::generateRecordType(const NodePtr &n) {
         // so we need to call decorate on it
         std::string decoratedNameAt = decorate(n->nameAt(i));
         os_ << "        " << decoratedNameAt << "(";
-        if (!noUnion_ && n->leafAt(i)->type() == avro::AVRO_UNION) {
-            // the nameAt(i) does not take c++ reserved words into account
-            // so we need to call decorate on it
-            os_ << decoratedNameAt << "_t";
-        } else {
-            os_ << types[i];
-        }
+        os_ << types[i];
         os_ << "())";
         if (i != (c - 1)) {
             os_ << ',';
@@ -305,17 +307,6 @@ void makeCanonical(string &s, bool foldCase) {
     }
 }
 
-string CodeGen::unionName() {
-    string s = schemaFile_;
-    string::size_type n = s.find_last_of("/\\");
-    if (n != string::npos) {
-        s = s.substr(n);
-    }
-    makeCanonical(s, false);
-
-    return s + "_Union__" + boost::lexical_cast<string>(unionNumber_++) + "__";
-}
-
 static void generateGetterAndSetter(ostream &os,
                                     const string &structName, const string &type, const string &name,
                                     size_t idx) {
@@ -323,12 +314,21 @@ static void generateGetterAndSetter(ostream &os,
 
     os << "inline\n";
 
-    os << type << sn << "get_" << name << "() const {\n"
+    os << "const " << type << "&" << sn << "get_" << name << "() const {\n"
        << "    if (idx_ != " << idx << ") {\n"
        << "        throw avro::Exception(\"Invalid type for "
-       << "union\");\n"
+       << "union " << structName << "\");\n"
        << "    }\n"
-       << "    return " << ANY_NS << "::any_cast<" << type << " >(value_);\n"
+       << "    return *std::any_cast<" << type << " >(&value_);\n"
+       << "}\n\n";
+
+    os << "inline\n"
+       << type << "&" << sn << "get_" << name << "() {\n"
+       << "    if (idx_ != " << idx << ") {\n"
+       << "        throw avro::Exception(\"Invalid type for "
+       << "union " << structName << "\");\n"
+       << "    }\n"
+       << "    return *std::any_cast<" << type << " >(&value_);\n"
        << "}\n\n";
 
     os << "inline\n"
@@ -337,6 +337,13 @@ static void generateGetterAndSetter(ostream &os,
        << "    idx_ = " << idx << ";\n"
        << "    value_ = v;\n"
        << "}\n\n";
+
+    os << "inline\n"
+       << "void" << sn << "set_" << name
+       << "(" << type << "&& v) {\n"
+       << "    idx_ = " << idx << ";\n"
+       << "    value_ = std::move(v);\n"
+       << "}\n\n";
 }
 
 static void generateConstructor(ostream &os,
@@ -380,14 +387,43 @@ string CodeGen::generateUnionType(const NodePtr &n) {
         return done[n];
     }
 
-    auto result = unionName();
+    // re-use existing union types that have the exact same branches
+    if (const auto existingName = unionTracker_.getExistingUnionName(types); existingName.has_value()) {
+        return existingName.value();
+    }
+    const std::string result = unionTracker_.generateNewUnionName(types);
 
     os_ << "struct " << result << " {\n"
         << "private:\n"
         << "    size_t idx_;\n"
-        << "    " << ANY_NS << "::any value_;\n"
-        << "public:\n"
-        << "    size_t idx() const { return idx_; }\n";
+        << "    std::any value_;\n"
+        << "public:\n";
+
+    os_ << "    /** enum representing union branches as returned by the idx() function */\n"
+        << "    enum class Branch: size_t {\n";
+
+    // generate a enum that maps the branch name to the corresponding index (as returned by idx())
+    std::set<std::string> used_branch_names;
+    for (size_t i = 0; i < c; ++i) {
+        // escape reserved literals for c++
+        auto branch_name = decorate(names[i]);
+        // avoid rare collisions, e.g. someone might name their struct int_
+        if (used_branch_names.find(branch_name) != used_branch_names.end()) {
+            size_t postfix = 2;
+            std::string escaped_name = branch_name + "_" + std::to_string(postfix);
+            while (used_branch_names.find(escaped_name) != used_branch_names.end()) {
+                ++postfix;
+                escaped_name = branch_name + "_" + std::to_string(postfix);
+            }
+            branch_name = escaped_name;
+        }
+        os_ << "        " << branch_name << " = " << i << ",\n";
+        used_branch_names.insert(branch_name);
+    }
+    os_ << "    };\n";
+
+    os_ << "    size_t idx() const { return idx_; }\n";
+    os_ << "    Branch branch() const { return static_cast<Branch>(idx_); }\n";
 
     for (size_t i = 0; i < c; ++i) {
         const NodePtr &nn = n->leafAt(i);
@@ -397,14 +433,16 @@ string CodeGen::generateUnionType(const NodePtr &n) {
                 << "    }\n"
                 << "    void set_null() {\n"
                 << "        idx_ = " << i << ";\n"
-                << "        value_ = " << ANY_NS << "::any();\n"
+                << "        value_ = std::any();\n"
                 << "    }\n";
         } else {
             const string &type = types[i];
             const string &name = names[i];
-            os_ << "    " << type << " get_" << name << "() const;\n"
-                                                        "    void set_"
-                << name << "(const " << type << "& v);\n";
+            os_ << "    "
+                << "const " << type << "& get_" << name << "() const;\n"
+                << "    " << type << "& get_" << name << "();\n"
+                << "    void set_" << name << "(const " << type << "& v);\n"
+                << "    void set_" << name << "(" << type << "&& v);\n";
             pendingGettersAndSetters.emplace_back(result, type, name, i);
         }
     }
@@ -551,8 +589,22 @@ void CodeGen::generateRecordTraits(const NodePtr &n) {
     }
 
     string fn = fullname(decorate(n->name()));
-    os_ << "template<> struct codec_traits<" << fn << "> {\n"
-        << "    static void encode(Encoder& e, const " << fn << "& v) {\n";
+    os_ << "template<> struct codec_traits<" << fn << "> {\n";
+
+    if (c == 0) {
+        os_ << "    static void encode(Encoder&, const " << fn << "&) {}\n";
+        // ResolvingDecoder::fieldOrder mutates the state of the decoder, so if that decoder is
+        // passed in, we need to call the method even though it will return an empty vector.
+        os_ << "    static void decode(Decoder& d, " << fn << "&) {\n";
+        os_ << "        if (avro::ResolvingDecoder *rd = dynamic_cast<avro::ResolvingDecoder *>(&d)) {\n";
+        os_ << "            rd->fieldOrder();\n";
+        os_ << "        }\n";
+        os_ << "    }\n";
+        os_ << "};\n";
+        return;
+    }
+
+    os_ << "    static void encode(Encoder& e, const " << fn << "& v) {\n";
 
     for (size_t i = 0; i < c; ++i) {
         // the nameAt(i) does not take c++ reserved words into account
@@ -596,6 +648,11 @@ void CodeGen::generateRecordTraits(const NodePtr &n) {
 }
 
 void CodeGen::generateUnionTraits(const NodePtr &n) {
+    const string name = done[n];
+    const string fn = fullname(name);
+    if (unionTracker_.unionTraitsAlreadyGenerated(fn)) {
+        return;
+    }
     size_t c = n->leaves();
 
     for (size_t i = 0; i < c; ++i) {
@@ -603,9 +660,6 @@ void CodeGen::generateUnionTraits(const NodePtr &n) {
         generateTraits(nn);
     }
 
-    string name = done[n];
-    string fn = fullname(name);
-
     os_ << "template<> struct codec_traits<" << fn << "> {\n"
         << "    static void encode(Encoder& e, " << fn << " v) {\n"
         << "        e.encodeUnionIndex(v.idx());\n"
@@ -641,7 +695,7 @@ void CodeGen::generateUnionTraits(const NodePtr &n) {
             os_ << "            {\n"
                 << "                " << cppTypeOf(nn) << " vv;\n"
                 << "                avro::decode(d, vv);\n"
-                << "                v.set_" << cppNameOf(nn) << "(vv);\n"
+                << "                v.set_" << cppNameOf(nn) << "(std::move(vv));\n"
                 << "            }\n";
         }
         os_ << "            break;\n";
@@ -649,6 +703,8 @@ void CodeGen::generateUnionTraits(const NodePtr &n) {
     os_ << "        }\n"
         << "    }\n"
         << "};\n\n";
+
+    unionTracker_.setTraitsGenerated(fn);
 }
 
 void CodeGen::generateTraits(const NodePtr &n) {
@@ -681,6 +737,43 @@ void CodeGen::generateTraits(const NodePtr &n) {
     }
 }
 
+void CodeGen::generateDocComment(const NodePtr &n, const char *indent) {
+    if (!n->getDoc().empty()) {
+        std::vector<std::string> lines;
+        {
+            const std::string &doc = n->getDoc();
+            size_t pos = 0;
+            size_t found;
+            while ((found = doc.find('\n', pos)) != std::string::npos) {
+                lines.push_back(doc.substr(pos, found - pos));
+                pos = found + 1;
+            }
+            if (pos < doc.size()) {
+                lines.push_back(doc.substr(pos));
+            }
+        }
+        for (auto &line : lines) {
+            line.erase(std::remove(line.begin(), line.end(), '\r'), line.end());
+
+            if (line.empty()) {
+                os_ << indent << "//\n";
+            } else {
+                // If a comment line ends with a backslash or backslash and whitespace,
+                // avoid generating code which will generate multi-line comment warnings
+                // on GCC. We can't just append whitespace here as escaped newlines ignore
+                // trailing whitespace.
+                auto lastBackslash = std::find(line.rbegin(), line.rend(), '\\');
+                auto lastNonWs = std::find_if(line.rbegin(), line.rend(), [](char c) { return !std::isspace(static_cast<int>(c)); });
+                // Note: lastBackslash <= lastNonWs because the iterators are reversed, "less" is later in the string.
+                if (lastBackslash != line.rend() && lastBackslash <= lastNonWs) {
+                    line.append("(backslash)");
+                }
+                os_ << indent << "// " << line << "\n";
+            }
+        }
+    }
+}
+
 void CodeGen::emitCopyright() {
     os_ << "/**\n"
            " * Licensed to the Apache Software Foundation (ASF) under one\n"
@@ -702,17 +795,22 @@ void CodeGen::emitCopyright() {
            " * See the License for the specific language governing "
            "permissions and\n"
            " * limitations under the License.\n"
-           " */\n\n\n";
+           " */\n\n";
+}
+
+void CodeGen::emitGeneratedWarning() {
+    os_ << "/* This code was generated by avrogencpp " << AVRO_VERSION << ". Do not edit.*/\n\n";
 }
 
 string CodeGen::guard() {
     string h = headerFile_;
     makeCanonical(h, true);
-    return h + "_" + lexical_cast<string>(random_()) + "__H_";
+    return h + "_" + std::to_string(random_()) + "_H";
 }
 
 void CodeGen::generate(const ValidSchema &schema) {
     emitCopyright();
+    emitGeneratedWarning();
 
     string h = guardString_.empty() ? guard() : guardString_;
 
@@ -720,24 +818,15 @@ void CodeGen::generate(const ValidSchema &schema) {
     os_ << "#define " << h << "\n\n\n";
 
     os_ << "#include <sstream>\n"
-#if __cplusplus >= 201703L
         << "#include <any>\n"
-#else
-        << "#include \"boost/any.hpp\"\n"
-#endif
+        << "#include <utility>\n"
         << "#include \"" << includePrefix_ << "Specific.hh\"\n"
         << "#include \"" << includePrefix_ << "Encoder.hh\"\n"
         << "#include \"" << includePrefix_ << "Decoder.hh\"\n"
         << "\n";
 
-    vector<string> nsVector;
     if (!ns_.empty()) {
-        boost::algorithm::split_regex(nsVector, ns_, boost::regex("::"));
-        for (vector<string>::const_iterator it =
-                 nsVector.begin();
-             it != nsVector.end(); ++it) {
-            os_ << "namespace " << *it << " {\n";
-        }
+        os_ << "namespace " << ns_ << " {\n";
         inNamespace_ = true;
     }
 
@@ -760,17 +849,11 @@ void CodeGen::generate(const ValidSchema &schema) {
 
     if (!ns_.empty()) {
         inNamespace_ = false;
-        for (vector<string>::const_iterator it =
-                 nsVector.begin();
-             it != nsVector.end(); ++it) {
-            os_ << "}\n";
-        }
+        os_ << "}\n";
     }
 
     os_ << "namespace avro {\n";
 
-    unionNumber_ = 0;
-
     generateTraits(root);
 
     os_ << "}\n";
@@ -779,19 +862,24 @@ void CodeGen::generate(const ValidSchema &schema) {
     os_.flush();
 }
 
-namespace po = boost::program_options;
-
 static string readGuard(const string &filename) {
     std::ifstream ifs(filename.c_str());
     string buf;
     string candidate;
     while (std::getline(ifs, buf)) {
-        boost::algorithm::trim(buf);
+        if (!buf.empty()) {
+            size_t start = 0, end = buf.length();
+            while (start < end && std::isspace(buf[start], std::locale::classic())) start++;
+            while (start < end && std::isspace(buf[end - 1], std::locale::classic())) end--;
+            if (start > 0 || end < buf.length()) {
+                buf = buf.substr(start, end - start);
+            }
+        }
         if (candidate.empty()) {
-            if (boost::algorithm::starts_with(buf, "#ifndef ")) {
+            if (buf.compare(0, 8, "#ifndef ") == 0) {
                 candidate = buf.substr(8);
             }
-        } else if (boost::algorithm::starts_with(buf, "#define ")) {
+        } else if (buf.compare(0, 8, "#define ") == 0) {
             if (candidate == buf.substr(8)) {
                 break;
             }
@@ -802,41 +890,107 @@ static string readGuard(const string &filename) {
     return candidate;
 }
 
+struct ProgramOptions {
+    bool helpRequested = false;
+    bool versionRequested = false;
+    bool noUnionTypedef = false;
+    std::string includePrefix = "avro";
+    std::string nameSpace;
+    std::string inputFile;
+    std::string outputFile;
+};
+
+static void printUsage() {
+    std::cout << "Allowed options:\n"
+              << "  -h [ --help ]                       produce help message\n"
+              << "  -V [ --version ]                    produce version information\n"
+              << "  -p [ --include-prefix ] arg (=avro) prefix for include headers, - for none, default: avro\n"
+              << "  -U [ --no-union-typedef ]           do not generate typedefs for unions in records\n"
+              << "  -n [ --namespace ] arg              set namespace for generated code\n"
+              << "  -i [ --input ] arg                  input file\n"
+              << "  -o [ --output ] arg                 output file to generate\n";
+}
+
+static bool parseArgs(int argc, char **argv, ProgramOptions &opts) {
+    for (int i = 1; i < argc; ++i) {
+        std::string arg = argv[i];
+
+        if (arg == "-h" || arg == "--help") {
+            opts.helpRequested = true;
+            return true;
+        }
+
+        if (arg == "-V" || arg == "--version") {
+            opts.versionRequested = true;
+            return true;
+        }
+
+        if (arg == "-U" || arg == "--no-union-typedef") {
+            opts.noUnionTypedef = true;
+            continue;
+        }
+
+        if (arg == "-p" || arg == "--include-prefix") {
+            if (i + 1 < argc) {
+                opts.includePrefix = argv[++i];
+                continue;
+            }
+        } else if (arg == "-n" || arg == "--namespace") {
+            if (i + 1 < argc) {
+                opts.nameSpace = argv[++i];
+                continue;
+            }
+        } else if (arg == "-i" || arg == "--input") {
+            if (i + 1 < argc) {
+                opts.inputFile = argv[++i];
+                continue;
+            }
+        } else if (arg == "-o" || arg == "--output") {
+            if (i + 1 < argc) {
+                opts.outputFile = argv[++i];
+                continue;
+            }
+        } else {
+            std::cerr << "Unknown option: " << arg << std::endl;
+            return false;
+        }
+
+        std::cerr << "Missing value for option: " << arg << std::endl;
+        return false;
+    }
+
+    return true;
+}
+
 int main(int argc, char **argv) {
-    const string NS("namespace");
-    const string OUT_FILE("output");
-    const string IN_FILE("input");
-    const string INCLUDE_PREFIX("include-prefix");
-    const string NO_UNION_TYPEDEF("no-union-typedef");
+    ProgramOptions opts;
+    if (!parseArgs(argc, argv, opts)) {
+        printUsage();
+        return 1;
+    }
 
-    po::options_description desc("Allowed options");
-    desc.add_options()("help,h", "produce help message")("version,V", "produce version information")("include-prefix,p", po::value<string>()->default_value("avro"),
-                                                         "prefix for include headers, - for none, default: avro")("no-union-typedef,U", "do not generate typedefs for unions in records")("namespace,n", po::value<string>(), "set namespace for generated code")("input,i", po::value<string>(), "input file")("output,o", po::value<string>(), "output file to generate");
-
-    po::variables_map vm;
-    po::store(po::parse_command_line(argc, argv, desc), vm);
-    po::notify(vm);
-
-    if (vm.count("help")) {
-        std::cout << desc << std::endl;
+    if (opts.helpRequested) {
+        printUsage();
         return 0;
     }
 
-    if (vm.count("version")) {
+    if (opts.versionRequested) {
         std::cout << AVRO_VERSION << std::endl;
         return 0;
     }
 
-    if (vm.count(IN_FILE) == 0 || vm.count(OUT_FILE) == 0) {
-        std::cout << desc << std::endl;
+    if (opts.inputFile.empty() || opts.outputFile.empty()) {
+        std::cerr << "Input and output files are required.\n\n";
+        printUsage();
         return 1;
     }
 
-    string ns = vm.count(NS) > 0 ? vm[NS].as<string>() : string();
-    string outf = vm.count(OUT_FILE) > 0 ? vm[OUT_FILE].as<string>() : string();
-    string inf = vm.count(IN_FILE) > 0 ? vm[IN_FILE].as<string>() : string();
-    string incPrefix = vm[INCLUDE_PREFIX].as<string>();
-    bool noUnion = vm.count(NO_UNION_TYPEDEF) != 0;
+    std::string ns = opts.nameSpace;
+    std::string outf = opts.outputFile;
+    std::string inf = opts.inputFile;
+    std::string incPrefix = opts.includePrefix;
+    bool noUnion = opts.noUnionTypedef;
+
     if (incPrefix == "-") {
         incPrefix.clear();
     } else if (*incPrefix.rbegin() != '/') {
@@ -867,3 +1021,34 @@ int main(int argc, char **argv) {
         return 1;
     }
 }
+
+UnionCodeTracker::UnionCodeTracker(const std::string &schemaFile) : schemaFile_(schemaFile) {
+}
+
+std::optional<std::string> UnionCodeTracker::getExistingUnionName(const std::vector<std::string> &unionBranches) const {
+    if (const auto it = unionBranchNameMapping_.find(unionBranches); it != unionBranchNameMapping_.end()) {
+        return it->second;
+    }
+    return std::nullopt;
+}
+
+std::string UnionCodeTracker::generateNewUnionName(const std::vector<std::string> &unionBranches) {
+    string s = schemaFile_;
+    string::size_type n = s.find_last_of("/\\");
+    if (n != string::npos) {
+        s = s.substr(n);
+    }
+    makeCanonical(s, false);
+
+    std::string result = s + "_Union__" + std::to_string(unionNumber_++) + "__";
+    unionBranchNameMapping_.emplace(unionBranches, result);
+    return result;
+}
+
+bool UnionCodeTracker::unionTraitsAlreadyGenerated(const std::string &unionClassName) const {
+    return generatedUnionTraits_.find(unionClassName) != generatedUnionTraits_.end();
+}
+
+void UnionCodeTracker::setTraitsGenerated(const std::string &unionClassName) {
+    generatedUnionTraits_.insert(unionClassName);
+}
