@@ -17,7 +17,6 @@
  */
 
 #include <algorithm>
-#include <boost/math/special_functions/fpclassify.hpp>
 #include <map>
 #include <memory>
 #include <string>
@@ -539,7 +538,7 @@ void JsonEncoder<P, F>::encodeFloat(float f) {
         out_.encodeString("Infinity");
     } else if (-f == std::numeric_limits<float>::infinity()) {
         out_.encodeString("-Infinity");
-    } else if (boost::math::isnan(f)) {
+    } else if (std::isnan(f)) {
         out_.encodeString("NaN");
     } else {
         out_.encodeNumber(f);
@@ -553,7 +552,7 @@ void JsonEncoder<P, F>::encodeDouble(double d) {
         out_.encodeString("Infinity");
     } else if (-d == std::numeric_limits<double>::infinity()) {
         out_.encodeString("-Infinity");
-    } else if (boost::math::isnan(d)) {
+    } else if (std::isnan(d)) {
         out_.encodeString("NaN");
     } else {
         out_.encodeNumber(d);
